]> git.ipfire.org Git - thirdparty/mdadm.git/blobdiff - bitmap.c
imsm: add support for checkpointing via 'curr_migr_unit'
[thirdparty/mdadm.git] / bitmap.c
index 8210278f356910fda95c9ac741475d8afcd80c17..088e37d0ad79f90bcf02c31684b2a2c9aeb6f8b4 100644 (file)
--- a/bitmap.c
+++ b/bitmap.c
@@ -20,8 +20,6 @@
 
 #include "mdadm.h"
 
-#define min(a,b) (((a) < (b)) ? (a) : (b))
-
 inline void sb_le_to_cpu(bitmap_super_t *sb)
 {
        sb->magic = __le32_to_cpu(sb->magic);
@@ -33,6 +31,7 @@ inline void sb_le_to_cpu(bitmap_super_t *sb)
        sb->chunksize = __le32_to_cpu(sb->chunksize);
        sb->daemon_sleep = __le32_to_cpu(sb->daemon_sleep);
        sb->sync_size = __le64_to_cpu(sb->sync_size);
+       sb->write_behind = __le32_to_cpu(sb->write_behind);
 }
 
 inline void sb_cpu_to_le(bitmap_super_t *sb)
@@ -114,6 +113,15 @@ unsigned long long bitmap_bits(unsigned long long array_size,
        return (array_size * 512 + chunksize - 1) / chunksize;
 }
 
+unsigned long bitmap_sectors(struct bitmap_super_s *bsb)
+{
+       unsigned long long bits = bitmap_bits(__le64_to_cpu(bsb->sync_size),
+                                             __le32_to_cpu(bsb->chunksize));
+       int bits_per_sector = 8*512;
+       return (bits + bits_per_sector - 1) / bits_per_sector;
+}
+
+
 bitmap_info_t *bitmap_fd_read(int fd, int brief)
 {
        /* Note: fd might be open O_DIRECT, so we must be
@@ -121,11 +129,13 @@ bitmap_info_t *bitmap_fd_read(int fd, int brief)
         */
        unsigned long long total_bits = 0, read_bits = 0, dirty_bits = 0;
        bitmap_info_t *info;
-       char *buf, *unaligned;
+       void *buf;
        int n, skip;
 
-       unaligned = malloc(8192*2);
-       buf = (char*) ((unsigned long)unaligned | 8191)+1;
+       if (posix_memalign(&buf, 512, 8192) != 0) {
+               fprintf(stderr, Name ": failed to allocate 8192 bytes\n");
+               return NULL;
+       }
        n = read(fd, buf, 8192);
 
        info = malloc(sizeof(*info));
@@ -144,15 +154,14 @@ bitmap_info_t *bitmap_fd_read(int fd, int brief)
                fprintf(stderr, Name ": failed to read superblock of bitmap "
                        "file: %s\n", strerror(errno));
                free(info);
-               free(unaligned);
                return NULL;
        }
        memcpy(&info->sb, buf, sizeof(info->sb));
        skip = sizeof(info->sb);
 
        sb_le_to_cpu(&info->sb); /* convert superblock to CPU byte ordering */
-       
-       if (brief || info->sb.sync_size == 0)
+
+       if (brief || info->sb.sync_size == 0 || info->sb.chunksize == 0)
                goto out;
 
        /* read the rest of the file counting total bits and dirty bits --
@@ -218,8 +227,8 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
                if (!st) {
                        /* just look at device... */
                        lseek(fd, 0, 0);
-               } else {        
-                       st->ss->locate_bitmap(st, fd, NULL);
+               } else {
+                       st->ss->locate_bitmap(st, fd);
                }
                ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
                *stp = st;
@@ -259,6 +268,8 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
        bitmap_info_t *info;
        int rv = 1;
        char buf[64];
+       int swap;
+       __u32 uuid32[4];
 
        info = bitmap_file_read(filename, brief, &st);
        if (!info)
@@ -278,19 +289,28 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
        }
 
        rv = 0;
-       if (st && st->ss->swapuuid) {
-       printf("            UUID : %08x.%08x.%08x.%08x\n",
-                                       swapl(*(__u32 *)(sb->uuid+0)),
-                                       swapl(*(__u32 *)(sb->uuid+4)),
-                                       swapl(*(__u32 *)(sb->uuid+8)),
-                                       swapl(*(__u32 *)(sb->uuid+12)));
-       } else {
-       printf("            UUID : %08x.%08x.%08x.%08x\n",
-                                       *(__u32 *)(sb->uuid+0),
-                                       *(__u32 *)(sb->uuid+4),
-                                       *(__u32 *)(sb->uuid+8),
-                                       *(__u32 *)(sb->uuid+12));
-       }
+       if (st)
+               swap = st->ss->swapuuid;
+       else
+#if __BYTE_ORDER == BIG_ENDIAN
+               swap = 0;
+#else
+               swap = 1;
+#endif
+       memcpy(uuid32, sb->uuid, 16);
+       if (swap)
+               printf("            UUID : %08x:%08x:%08x:%08x\n",
+                      swapl(uuid32[0]),
+                      swapl(uuid32[1]),
+                      swapl(uuid32[2]),
+                      swapl(uuid32[3]));
+       else
+               printf("            UUID : %08x:%08x:%08x:%08x\n",
+                      uuid32[0],
+                      uuid32[1],
+                      uuid32[2],
+                      uuid32[3]);
+
        printf("          Events : %llu\n", (unsigned long long)sb->events);
        printf("  Events Cleared : %llu\n", (unsigned long long)sb->events_cleared);
        printf("           State : %s\n", bitmap_state(sb->state));
@@ -344,11 +364,12 @@ int CreateBitmap(char *filename, int force, char uuid[16],
        if (chunksize == UnSet) {
                /* We don't want more than 2^21 chunks, as 2^11 fill up one
                 * 4K page (2 bytes per chunk), and 2^10 address of those
-                * fill up a 4K indexing page.  2^20 might be safer...
+                * fill up a 4K indexing page.  2^20 might be safer, especially
+                * on 64bit hosts, so use that.
                 */
                chunksize = DEFAULT_BITMAP_CHUNK;
-               /* <<21 for 2^21 chunks, >>9 to convert bytes to sectors */
-               while (array_size > (chunksize << (21-9)))
+               /* <<20 for 2^20 chunks, >>9 to convert bytes to sectors */
+               while (array_size > (chunksize << (20-9)))
                        chunksize <<= 1;
        }
 
@@ -387,8 +408,9 @@ int CreateBitmap(char *filename, int force, char uuid[16],
                }
                bytes -= sizeof(block);
        }
-       
+
        rv = 0;
+       fflush(fp);
        /* make the file be the right size (well, to the nearest byte) */
        if (ftruncate(fileno(fp), filesize))
                perror("ftrunace");
@@ -399,16 +421,22 @@ out:
        return rv;
 }
 
-void bitmap_update_uuid(int fd, int *uuid)
+int bitmap_update_uuid(int fd, int *uuid, int swap)
 {
        struct bitmap_super_s bm;
-       lseek(fd, 0, 0);
+       if (lseek(fd, 0, 0) != 0)
+               return 1;
        if (read(fd, &bm, sizeof(bm)) != sizeof(bm))
-               return;
+               return 1;
        if (bm.magic != __cpu_to_le32(BITMAP_MAGIC))
-               return;
-       memcpy(bm.uuid, uuid, 16);
-       lseek(fd, 0, 0);
-       write(fd, &bm, sizeof(bm));
+               return 1;
+       copy_uuid(bm.uuid, uuid, swap);
+       if (lseek(fd, 0, 0) != 0)
+               return 2;
+       if (write(fd, &bm, sizeof(bm)) != sizeof(bm)) {
+               lseek(fd, 0, 0);
+               return 2;
+       }
        lseek(fd, 0, 0);
+       return 0;
 }